/*
  
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       CCM (bookKeeping.C)
  
  Written by: Damian J. J. Farnell
  
  Date:       Wed Sep 29 1999
  
  Purpose:    Determine and solve CCM LSUBm equations.
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: BOOKKEEPING.C 

  PURPOSE:     This module does all of the `bookkeeping' 
               for all of the other modules. Hence, all
               routines that are multiply used throughout
               the code are contained here. 

  ---------------------------------------------------

*/


#include<iostream.h>
#include<stdlib.h>
#include<stdlib.h>
#include<math.h>
#include<stdio.h>
#include<string.h>
#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

#include"../include/nr.h"
#include"../include/nrutil.h"
#include"../include/Constants.h"
#include"../include/Externals.h"
#include"../include/FunctionPrototypes.h"

void findUniqueDescriptorOfCluster(int order,
				   int **configuration,
				   int *descriptor){

  /* This routine gets a set of crystallographic lattice sites
     which define a cluster and returns a set of unique integer 
     numbers in `descriptor' which fully describe it.

     Note that this is its FUNDAMENTAL description, i.e., 
     of the how ever many possible symmetry equivalent 
     cluster there are of it, this `descriptor' has 
     the lowest numerical value.

     PARTS ARE:

     1) determine symmetry related equivalents and translate
     ...into the x>=0,y>=0,z>=0 part.

     2) determine the `spin' form of these equivalents

     3) turn this `spin' form into a unique set of integers

     4) find set of integers with lowest `numerical' value
     ...out of all the symmetry related clusters and return
     ...this in `descriptor'.

  */

  int i,j,k,l;
  int minX, minY, minZ;
  int **tempConfiguration;
  static int tempDescriptors[MAX_NUM_SYM][MAX_SPIN_LIST];
  static int minDescriptors[MAX_SPIN_LIST];
  int WORKING_OK,OK; 

  tempConfiguration=matrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=LatticeSymmetries.number;i++) {

    // find symmetry related clusters

    symmetryGroup(i,order,configuration,tempConfiguration);

    if (Control.output >4)
      for(j=1;j<=order;j++)
	printf("transformed lattice site: (%d, %d, %d, %d)\n",
	       tempConfiguration[j][1],
	       tempConfiguration[j][2],
	       tempConfiguration[j][3],
	       tempConfiguration[j][4]);
    
    // translate into correct quadrant, x>=0,y>=0,z>=0
    
    minX=10000000;
    minY=10000000;
    minZ=10000000;

    for(j=1;j<=order;j++) {
      if (tempConfiguration[j][1]<minX)
	minX=tempConfiguration[j][1];
      if (tempConfiguration[j][2]<minY)
	minY=tempConfiguration[j][2];
      if (tempConfiguration[j][3]<minZ)
	minZ=tempConfiguration[j][3];
    }

    for(j=1;j<=order;j++) {
      tempConfiguration[j][1]=tempConfiguration[j][1]-minX;
      tempConfiguration[j][2]=tempConfiguration[j][2]-minY;
      tempConfiguration[j][3]=tempConfiguration[j][3]-minZ;
    }

    if (Control.output >4)
      for(j=1;j<=order;j++)
	printf("translated lattice site: (%d, %d, %d, %d)\n",
	       tempConfiguration[j][1],
	       tempConfiguration[j][2],
	       tempConfiguration[j][3],
	       tempConfiguration[j][4]);

    // now find the descriptor
    
    findConfigurationToDescriptor(tempDescriptors[i],tempConfiguration,order);

    if (Control.output > 4) {
      printf("Temp. Descriptor(%d) = ",i);
      for(j=1;j<MAX_SPIN_LIST;j++)
	cout << tempDescriptors[i][j] << " ";
      cout << endl;
      
    }
  }

  // now find the smallest numerical valued one
    
  for(k=1;k<MAX_SPIN_LIST;k++)
    minDescriptors[k]=1000000000;
  
  WORKING_OK=0;

  for(i=1;i<=LatticeSymmetries.number;i++) {

    OK=0;
    
    for(k=(MAX_SPIN_LIST-1);k>0;k--) {
      if (tempDescriptors[i][k] < minDescriptors[k]) {
	OK=1;
	break;
      } else if (tempDescriptors[i][k] > minDescriptors[k]) {
	break;
      }
    }

    if (OK) {
      for(k=1;k<MAX_SPIN_LIST;k++)
	minDescriptors[k]=tempDescriptors[i][k];
      WORKING_OK=1;
    }
  }

  if (!WORKING_OK) {
    cout << "Can't find a reasonable descriptor of this cluster!\n"
      "Exiting...\n\n";
    exit(1);
  } else {
    for(k=1;k<MAX_SPIN_LIST;k++)
      descriptor[k]=minDescriptors[k];
  }
   
  if (Control.output > 6) {
    printf("Minimum Descriptors = ");
    for(j=1;j<MAX_SPIN_LIST;j++)
      cout << descriptor[j] << " ";
    cout << endl;
  }
  free_matrix(tempConfiguration,0,MAX_LSUBM+1,0,5);
 
}

void symmetryGroup(int switchVariable, 
		   int order,
		   int **configuration,
		   int **tempConfiguration)
{

  int i;

  double newx, newy, newz;
  double x,y,z;

  // symmorphic point group symetries entered here only =>
  // triangular lattice not-yet implemented, to be done.

  if (switchVariable > 48 || switchVariable < 1) {
    cout << "Switch Variable out of range " << switchVariable << 
      "...\nExiting!\n\n";
    exit(1);
  }

  for(i=1;i<=order;i++) {
    
    findCartesiansFromGeneralisedPosition(x, y, z, configuration[i]);
    
    newx=x*LatticeSymmetries.R[switchVariable][1][1]+
      y*LatticeSymmetries.R[switchVariable][1][2]+
      z*LatticeSymmetries.R[switchVariable][1][3]+
      LatticeSymmetries.x[switchVariable];
    
    newy=x*LatticeSymmetries.R[switchVariable][2][1]+
      y*LatticeSymmetries.R[switchVariable][2][2]+
      z*LatticeSymmetries.R[switchVariable][2][3]+
      LatticeSymmetries.y[switchVariable];
    
    newz=x*LatticeSymmetries.R[switchVariable][3][1]+
      y*LatticeSymmetries.R[switchVariable][3][2]+
      z*LatticeSymmetries.R[switchVariable][3][3]+
      LatticeSymmetries.z[switchVariable];
    
    findGeneralisedPositionFromCartesians(newx,newy,newz,tempConfiguration[i]);
  }
}

void findGeneralisedPositionFromCartesians(double &x, 
					   double &y, 
					   double &z, 
					   int *configuration)
{
  
  int index1, index2, index3, index4;
  int i,j,k,l,found;

  double **m, **tempm;
  double *s,distance;
  int *ind;

  double d;

  ind=ivector(0,5);
  s=dvector(0,5);
  m=dmatrix(0,5,0,5); 
  tempm=dmatrix(0,5,0,5); 

  // this is a rather complicated way of doing this though quite
  // quick -- see linear search below.
  
  // for simpler lattices this whole piece of code can be replaced
  // by a far simpler algroithm.
  
  // note that this code is implemented in 2D only here.
  
  m[1][1]=BravaisLattice.a[1];
  m[2][1]=BravaisLattice.a[2];
  m[3][1]=BravaisLattice.a[3];

  m[1][2]=BravaisLattice.b[1];
  m[2][2]=BravaisLattice.b[2];
  m[3][2]=BravaisLattice.b[3];

  m[1][3]=BravaisLattice.c[1];
  m[2][3]=BravaisLattice.c[2];
  m[3][3]=BravaisLattice.c[3];

  found=0;

  index4=-1;
  for(i=1;i<=UnitCell.number;i++) {

    s[1]=x-UnitCell.x[i];
    s[2]=y-UnitCell.y[i];
    s[3]=z-UnitCell.z[i];

    for(j=1;j<=3;j++)
      for(k=1;k<=3;k++)
	tempm[j][k]=m[j][k];

    ludcmp(tempm,3,ind,&d);
    lubksb(tempm,3,ind,s);

    // check if the values for s[] are integers.

    if (s[1] > 0.001)
      index1=(int) (s[1]+0.001);
    else if (fabs(s[1]) <= 0.001)
      index1=0;
    else if (s[1] < 0.0) 
      index1=(int) (s[1]-0.001);

    if (s[2] > 0.001) 
      index2=(int) (s[2]+0.001);
    else if (fabs(s[2]) <= 0.001)
      index2=0; 
    else
      index2 = (int)(s[2]-0.001);

    if (s[3] > 0.001)
      index3=(int) (s[3]+0.001);
    else if (fabs(s[3]) <= 0.001)
      index3=0;
    else
      index3=(int) (s[3]-0.001);

    index4=i;

    distance=fabs(x-index1*BravaisLattice.a[1]-index2*BravaisLattice.b[1]
		  -index3*BravaisLattice.c[1]-UnitCell.x[index4]);
    distance+=fabs(y-index1*BravaisLattice.a[2]-index2*BravaisLattice.b[2]
		  -index3*BravaisLattice.c[2]-UnitCell.y[index4]);
    distance+=fabs(z-index1*BravaisLattice.a[3]-index2*BravaisLattice.b[3]
		  -index3*BravaisLattice.c[3]-UnitCell.z[index4]);
    if (distance < 0.001) {
      found=1;
      break;
    }
  }

  if (!found) { 
    cout << "Error in you findGeneralisedPositionFromCartesians " 
	 << index1 << " " << index2 <<
      " " << index3 << " " << index4 << endl;
    cout << x << " " << y << " " << z << endl;
    exit(0);
  }

  free_dmatrix(m,0,5,0,5);
  free_dmatrix(tempm,0,5,0,5);
  free_ivector(ind,0,5);
  free_dvector(s,0,5);

  configuration[1]=index1;
  configuration[2]=index2;
  configuration[3]=index3;
  configuration[4]=index4;
  

}
  
void findCartesiansFromGeneralisedPosition(double &x, 
					   double &y, 
					   double &z, 
					   int *configuration)
{

  // this routine finds the cartesian position from the 
  // `generalised' lattice position => bravais vector + primitive
  // cell vector

  x=BravaisLattice.a[1]*configuration[1]+
    BravaisLattice.b[1]*configuration[2]+
    BravaisLattice.c[1]*configuration[3]+
    UnitCell.x[configuration[4]];

  y=BravaisLattice.a[2]*configuration[1]+
    BravaisLattice.b[2]*configuration[2]+
    BravaisLattice.c[2]*configuration[3]+
    UnitCell.y[configuration[4]];

  z=BravaisLattice.a[3]*configuration[1]+
    BravaisLattice.b[3]*configuration[2]+
    BravaisLattice.c[3]*configuration[3]+
    UnitCell.z[configuration[4]];

}

void findConfigurationToDescriptor(int *descriptor,
				   int **config,
				   int order)
{

  int i,j,k,l,a;
  int length_of_chain;
  int index,l1,l2,l3;
  int tempNum, tempS;
  int block;
  static int checkBlock[MAX_BT+1][MAX_UNIT_CELL+1];

  // checkBlock makes sure that no more than 2*Params.m+1 spin
  // flips per site occur.
    
  for(i=0;i<=max_number_cells_BT;i++)
    for(j=0;j<=UnitCell.number;j++)
      checkBlock[i][j]=0;

  for(i=1;i<MAX_SPIN_LIST;i++) 
    descriptor[i]=0;
  
  // the first part is to determine a unique lattice index in the bounding
  // triangle for each spin flip operator in a given cluster. Each spin
  // flip is denoted by it bravais lattice vector + its number in the 
  // unit cell. 
  
  // each cell in the the bounding triangle is a unit cell.

  for(i=1;i<=order;i++) {

    if (Params.dim == 1) {
      index=config[i][1]+1;
      if (config[i][1] > (Params.btLength-1)) {
	// point lies outside the bounding triangle.
	for(i=1;i<MAX_SPIN_LIST;i++) 
	  descriptor[i]=999999;
	return;
      }
    } else if (Params.dim == 2) {

#if DEBUG
      if (config[i][2] < 0 || config[i][1] < 0) {
	cout << "Error (2D) findConfigurationToDescriptor...\n"
	     << "config[i][2] < 0 || config[i][1] < 0...\nExiting!!\n\n";
	exit(1);
      }
#endif

      // bounding triangle in 2D
      if ((abs(config[i][2])+abs(config[i][1])) > (Params.btLength-1)) {
	// point lies outside the bounding triangle.
#if 1
	for(i=1;i<MAX_SPIN_LIST;i++) 
	  descriptor[i]=999999;
#else
        descriptor[1]=999999;  /* faster? */
#endif
	return;
      }

      /*      index=Params.btLength*(Params.btLength+1)/2
	-(Params.btLength-config[i][2])*(Params.btLength-config[i][2]+1)/2
	+config[i][1]+1;
      */

      index=BTarray[config[i][1]][config[i][2]][0];

    } else {

#if DEBUG
      if (config[i][2] < 0 || config[i][1] < 0 || config[i][3] < 0) {
	cout << "Error (3D) findConfigurationToDescriptor...\n"
	     << "config[i][2] < 0 || config[i][1] < 0...\nExiting!!\n\n";
	exit(1);
      }
#endif

      // bounding triangle in 2D
      if ((abs(config[i][2])+abs(config[i][1])+abs(config[i][3])) 
	  > (Params.btLength-1)) {
	// point lies outside the bounding triangle.
	for(i=1;i<MAX_SPIN_LIST;i++) 
	  descriptor[i]=999999;
	return;
      }

      index=BTarray[config[i][1]][config[i][2]][config[i][3]];

      //      cout << "3D Bounding ``Triangle'' to be implemented...\n"
      // << "Exiting!\n\n";
      //      exit(0);
    }   

    checkBlock[index][config[i][4]]++;

    // at this point: 1 <= index <= max_numbers_cells_in_BT

    // we now have an unique site ``index'' for 1D-3D which completely
    // specifies the position of a given spin in the bounding triangle.
    // we now wish to encode this information in a given ``block'' 
    // with a given number ``tempNum'' in ``descriptor''.

    // note that BLOCK_LENGTH gives the numbers of ``spins'' the
    // each element of the descriptor describes. Normally set to
    // 20 such spins so that we can look at large values of the 
    // quantum spin number.

    index=(index-1)*UnitCell.number+config[i][4];

    // at this point: 1 <= index <= max_number_cells_BT*UnitCell.number
    // note that: 1 <= config[][4] <= UnitCell.number

    index--;

    block=(int) index/BLOCK_LENGTH;

    // note that we must check that 0 <= block < MAX_SPIN_LIST-1

#if DEBUG
    if (block >= MAX_SPIN_LIST-1) {
      cout << "`block' exceeds (MAX_SPIN_LIST-1)...\nExiting!\n\n";
      exit(1);
    }
#endif

    index=index-((int) block*BLOCK_LENGTH);
    
    // at this point: 0 <= index < BLOCK_LENGTH

    //    cout << "index = " << index << endl;

    tempS=((int) (2*Params.s+1.0000001)); 
    tempNum=1;
    for(j=1;j<=index;j++) {
#if DEBUG
      if (tempNum >= 2147438547/tempS) {
	cout << "tempNum >= biggest INTEGER possible...\nExiting!!\n\n";
	exit(1);
      }
#endif
      tempNum=tempNum*tempS;
    }
    //    cout << endl << tempNum << endl << endl;

    

    descriptor[block+1]+=tempNum;
  }  

#if DEBUG
  // Now do check that no configuration has more than 2*Params.s +1
  // spin flips at each site.

  for(i=0;i<=max_number_cells_BT;i++) {
    for(j=0;j<=UnitCell.number;j++) {
      if (checkBlock[i][j] >= (int) (2*Params.s +1.0000001)){ 
	/*	for(i=1;i<MAX_SPIN_LIST;i++) 
	  descriptor[i]=999999;
	  return; */
	cout << "Configuration has more than 2*Params.s spin flips..."
	     << "Exiting!\n\n";
	exit(1); 
      }
    }
  }
#endif
}

void printOutDescriptor(int *descriptor){
  int i,j,k,l,a;
  int length_of_chain;
  int index,l1,l2,l3;
  int tempNum, tempS;
  int block;

  printf("%d number descriptor:  ",MAX_SPIN_LIST-1);
  for(i=1;i<MAX_SPIN_LIST;i++) 
    printf("%d ",descriptor[i]);
  printf("\n");

}

void printOutClusterFromDescriptor(int *descriptor){

  int i,j,k,l,m;
  int index;
  int temp1, temp2, temp, tempS;
  long int divisors[BLOCK_LENGTH+3];

  int x,y,z;

  int cell;
  int checkBlock[400][MAX_UNIT_CELL];
  
  int flips, count;
  
  // checkBlock makes sure that no more than 2*Params.m+1 spin
  // flips per site occur.
  
  for(i=0;i<=max_number_cells_BT;i++)
    for(j=0;j<=UnitCell.number;j++)
      checkBlock[i][j]=0;

  // now find the divisors which give the numbers to put into checkBlock

  tempS=(int) (2*Params.s+1.000000001);
  divisors[1]=1;
  for(j=2;j<=BLOCK_LENGTH+1;j++) {
    divisors[j]=divisors[j-1]*tempS;
    if (divisors[j] > 0) {
    }

    //    cout << divisors[j] << " ";
  }
  //  cout << endl;

  for(i=1;i<MAX_SPIN_LIST;i++) {    
    if (descriptor[i] > 0) {
      temp1=descriptor[i];
      for(j=BLOCK_LENGTH;j>0;j--) {
	
	temp2=(int) (temp1/divisors[j]);     
	temp1=temp1-temp2*divisors[j];  
	
	index=j+(i-1)*BLOCK_LENGTH;

	// note that: 1 <= index < BLOCK_LENGTH*MAX_SPIN_LIST

	temp=index;
	index=(int) ((index-1)/UnitCell.number);
	cell=temp-UnitCell.number*index;
	index++;

	// 1 <= index <= (BLOCK_LENGTH*MAX_SPIN_LIST-1)/UnitCell.number 

	if (index > max_number_cells_BT && temp2 > 0) {
	  cout << "index (=" << index 
	       << ") > max_number_cells_BT (=" 
	       << max_number_cells_BT << ")...\nExiting...\n\n";
	  exit(1);
	} 

	if (temp2 > 0)
	  checkBlock[index][cell]+=temp2;

	if (temp2 > 0 && Control.output > 4)
	  cout << "j = " << j << " index = " << index << " cell = " << cell << 
	    " temp2 = " << temp2 << endl ;
	
      }
    }
  }

  // now count the number of spin flips

  flips=0;

  for(i=1;i<=max_number_cells_BT;i++) {
    for(j=1;j<=UnitCell.number;j++) {

      if (checkBlock[i][j] > (int) (2*Params.s)){ 
	/*	cout << "Configuration has more than 2*Params.s + 1 spin flips..."
	     << "Exiting!\n\n";
	     exit(1); */
      }

      if (checkBlock[i][j] > 0) {
	
	if (Params.dim==1) {
	  x=i-1;
	  y=0;
	  z=0;
	} else if (Params.dim==2) {  

	  count=0;
	  for (l=0;l<Params.btLength;l++) {
	    for(k=0;k<Params.btLength-l;k++) {
	      count++;
	      if (count == i)
		break;
	    }
	    if (count == i)
	      break;
	  }
	  x=k;
	  y=l;
	  z=0;
	} else {
	  x=lookUpTable[i][1];
	  y=lookUpTable[i][2];
	  z=lookUpTable[i][3];
	  //	  cout << "3D not done yet...\n\nExiting!\n\n";
	  // exit(1);
	}
      }
      
      for (m=1;m<=checkBlock[i][j];m++) {
	flips++;
	printf("lattice site(%d): (%d, %d, %d, %d)\n",flips,x,y,z,j);
      }
    }
  }
  cout << endl;
}


void printOutCluster(int length, int **configuration)
{	
  int l; 
  
  for(l=1;l<=length;l++) {
    printf("lattice site(%d): (%d, %d, %d, %d)\n",
	   l,
	   configuration[l][1],
	   configuration[l][2],
	   configuration[l][3],
	   configuration[l][4]);
  }
}

int findClusterFromDescriptor(int *descriptor, int **configuration){

  int i,j,k,l,m;
  int index;
  int temp1, temp2, temp, tempS;
  long int divisors[BLOCK_LENGTH+3];

  int x,y,z;

  int cell;
  int checkBlock[400][MAX_UNIT_CELL];
  
  int flips, count;
  
  // checkBlock makes sure that no more than 2*Params.m+1 spin
  // flips per site occur.
  
  for(i=0;i<=max_number_cells_BT;i++)
    for(j=0;j<=UnitCell.number;j++)
      checkBlock[i][j]=0;

  // now find the divisors which give the numbers to put into checkBlock

  tempS=(int) (2*Params.s+1.000000001);
  divisors[1]=1;
  for(j=2;j<=BLOCK_LENGTH+1;j++) {
    divisors[j]=divisors[j-1]*tempS;
    //    cout << divisors[j] << " ";
  }
  //  cout << endl;

  for(i=1;i<MAX_SPIN_LIST;i++) {    
    if (descriptor[i] > 0) {
      temp1=descriptor[i];
      for(j=BLOCK_LENGTH;j>0;j--) {
	
	temp2=(int) (temp1/divisors[j]);     
	temp1=temp1-temp2*divisors[j];  
	
	index=j+(i-1)*BLOCK_LENGTH;

	// note that: 1 <= index < BLOCK_LENGTH*MAX_SPIN_LIST

	temp=index;
	index=(int) ((index-1)/UnitCell.number);
	cell=temp-UnitCell.number*index;
	index++;

	// 1 <= index <= (BLOCK_LENGTH*MAX_SPIN_LIST-1)/UnitCell.number 

	if (index > max_number_cells_BT && temp2 > 0) {
	  cout << "index (=" << index 
	       << ") > max_number_cells_BT (=" 
	       << max_number_cells_BT << ")...\nExiting...\n\n";
	  exit(1);
	} 

	if (temp2 > 0)
	  checkBlock[index][cell]+=temp2;

	if (temp2 > 0 && Control.output > 4)
	  cout << "j = " << j << " index = " << index << " cell = " << cell << 
	    " temp2 = " << temp2 << endl ;
	
      }
    }
  }

  // now count the number of spin flips

  flips=0;

  for(i=1;i<=max_number_cells_BT;i++) {
    for(j=1;j<=UnitCell.number;j++) {

      if (checkBlock[i][j] > (int) (2*Params.s)){ 
	cout << "Configuration has more than 2*Params.s + 1 spin flips..."
	     << "Exiting!\n\n";
	exit(1);
      }

      if (checkBlock[i][j] > 0) {
	
	if (Params.dim==1) {
	  x=i-1;
	  y=0;
	  z=0;
	} else if (Params.dim==2) {  

	  /*	  count=0;
		  for (l=0;l<Params.btLength;l++) {
		  for(k=0;k<Params.btLength-l;k++) {
		  count++;
		  if (count == i)
		  break;
		  }
		  if (count == i)
		  break;
		  }*/

	  x=lookUpTable[i][1];
	  y=lookUpTable[i][2];
	  
	  //	  x=k;
	  //	  y=l;
	  z=0;
	} else {
	  x=lookUpTable[i][1];
	  y=lookUpTable[i][2];
	  z=lookUpTable[i][3];
	}
      }
      
      for (m=1;m<=checkBlock[i][j];m++) {
	flips++;
	configuration[flips][1]=x;
	configuration[flips][2]=y;
	configuration[flips][3]=z;
	configuration[flips][4]=j;	  
      }
    }
  }
  return flips;
}




int findDescriptorIndex(int *descriptor,
			int numberElements) {

  int tempLogicFlag;
  int position;
  int i,j;
  int tempPosition1;
  int tempPosition2;
  int middle;
  int switchFlag;

  if (numberElements < 1)
    return 0;

  tempLogicFlag=1;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (descriptor[i]!=999999) {
      tempLogicFlag=0;
      break;
    }
  }

  if (tempLogicFlag)
    return 0;

  position=0;

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  tempPosition1=1;
  tempPosition2=numberElements;

  // first check that the end elements do not match  `descriptor'

  tempLogicFlag=0;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (spinConfigurations[tempPosition1].descriptor[i]!=descriptor[i]) {
      tempLogicFlag=1;
      break;
    }
  }
  
  if (!tempLogicFlag)
    position=tempPosition1;
  
  if (tempLogicFlag) {
    tempLogicFlag=0;
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (spinConfigurations[tempPosition2].descriptor[i]!=descriptor[i]) {
	tempLogicFlag=1;
	break;
      }
    }
    if (!tempLogicFlag)
      position=tempPosition2;
  }

  // now check that elements in the `middle' do not match `descriptor'

  if (tempLogicFlag) {
    do {

      middle=(tempPosition1+tempPosition2)/2;      
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
	if (spinConfigurations[middle].descriptor[i]!=descriptor[i]) {
	  if (spinConfigurations[middle].descriptor[i] > descriptor[i])
	    switchFlag=1;
	  else
	    switchFlag=0;
	  tempLogicFlag=1;
	  break;
	}
      }      

      if (!tempLogicFlag) {
	position=middle;
	tempPosition2=tempPosition1;
      } else {
	if (switchFlag)
	  tempPosition2=middle;
	else
	  tempPosition1=middle;
      }

    } while ((tempPosition2-tempPosition1) > 1);
  }
		       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {

    tempLogicFlag=0;
    
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (spinConfigurations[j].descriptor[i]!=descriptor[i]) {
	tempLogicFlag=1;
	break;
      }
    }

    if (!tempLogicFlag) {
      position=j;
      break;
    }
    } 
  */

  return position;

}

void binary(int index, int length, int *spin)
{
  
  int i,y,dummy;    
  dummy=index;
  for(i=1;i<=length;i++) {
    y= dummy/2;
    spin[i]=dummy-2*y;
    dummy=y;
  }
}

void trimary(int index1, int index2, int index3, int length, int *spin)
{
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1;
  dummy2=index2;
  dummy3=index3;

  if (length > 20) { 
    
    for(i=21;i<=length;i++) {
      y= dummy1/3;
      spin[i]=dummy1-3*y;
      dummy1=y;
    }

    for(i=11;i<=20;i++) {
      y= dummy2/3;
      spin[i]=dummy2-3*y;
      dummy2=y;
    }

    for(i=1;i<=10;i++) {
      y= dummy3/3;
      spin[i]=dummy3-3*y;
      dummy3=y;
    }

  } else if (length > 10) {
    
    for(i=11;i<=length;i++) {
      y= dummy2/3;
      spin[i]=dummy2-3*y;
      dummy2=y;
    }
    
    for(i=1;i<=10;i++) {
      y= dummy3/3;
      spin[i]=dummy3-3*y;
      dummy3=y;
    } 

  } else {

    for(i=1;i<=length;i++) {
      y= dummy3/3;
      spin[i]=dummy3-3*y;
      dummy3=y;
    } 
  }
}


void quartary(int index1, int index2, int index3, int length, int *spin)
{
  
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1;
  dummy2=index2;
  dummy3=index3;

  if (length > 20) { 
    
    for(i=21;i<=length;i++) {
      y= dummy1/4;
      spin[i]=dummy1-4*y;
      dummy1=y;
    }

    for(i=11;i<=20;i++) {
      y= dummy2/4;
      spin[i]=dummy2-4*y;
      dummy2=y;
    }

    for(i=1;i<=10;i++) {
      y= dummy3/4;
      spin[i]=dummy3-4*y;
      dummy3=y;
    }

  } else if (length > 10) {
    
    for(i=11;i<=length;i++) {
      y= dummy2/4;
      spin[i]=dummy2-4*y;
      dummy2=y;
    }
    
    for(i=1;i<=10;i++) {
      y= dummy3/4;
      spin[i]=dummy3-4*y;
      dummy3=y;
    } 

  } else {

    for(i=1;i<=length;i++) {
      y= dummy3/4;
      spin[i]=dummy3-4*y;
      dummy3=y;
    } 
  }
}

int findDescriptorIndex2(int *descriptor,
                        int numberElements) {

  int tempLogicFlag;
  int position;
  int i,j;
  int tempPosition1;
  int tempPosition2;
  int middle;
  int switchFlag;

  if (numberElements < 1)
    return 0;

  tempLogicFlag=1;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (descriptor[i]!=999999) {
      tempLogicFlag=0;
      break;
    }
  }

  if (tempLogicFlag)
    return 0;

  position=0;

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  tempPosition1=1;
  tempPosition2=numberElements;

  // first check that the end elements do not match  `descriptor'

  tempLogicFlag=0;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (latticeAnimals[tempPosition1].descriptor[i]!=descriptor[i]) {
      tempLogicFlag=1;
      break;
    }
  }
  
  if (!tempLogicFlag)
    position=tempPosition1;
  
  if (tempLogicFlag) {
    tempLogicFlag=0;
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[tempPosition2].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }
    if (!tempLogicFlag)
      position=tempPosition2;
  }

  // now check that elements in the `middle' do not match `descriptor'

  if (tempLogicFlag) {
    do {

      middle=(tempPosition1+tempPosition2)/2;      
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
        if (latticeAnimals[middle].descriptor[i]!=descriptor[i]) {
          if (latticeAnimals[middle].descriptor[i] > descriptor[i])
            switchFlag=1;
          else
            switchFlag=0;
          tempLogicFlag=1;
          break;
        }
      }      

      if (!tempLogicFlag) {
        position=middle;
        tempPosition2=tempPosition1;
      } else {
        if (switchFlag)
          tempPosition2=middle;
        else
          tempPosition1=middle;
      }

    } while ((tempPosition2-tempPosition1) > 1);
  }
                       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {

    tempLogicFlag=0;
    
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[j].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }

    if (!tempLogicFlag) {
      position=j;
      break;
    }
    } 
  */

  return position;

}




